/*	Basic feed script loading all data from a local CSV file.

 	Copyright (c) 2021 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
 */

import * as feed from "system_lib/Feed";
import {SimpleFile} from "system/SimpleFile";
import {FeedDetailsSpec, FeedListSpec, ListData} from "system_lib/Feed";
import {field} from "system_lib/Metadata";

/**
 * Data read from CSV file.
 */
interface CSVData {
	startTime: string;
	endTime: string;
	firstName: string;
	lastName: string;
	city: string;
	company: string;
}

/**
 * Data provided to Blocks describing each feed item.
 */
class Item {

	@field("Given name")
	readonly firstName: string;

	@field("Surname")
	readonly lastName: string;

	@field("The company the fisitor work for")
	readonly company: string;

	@field("The city the visitor is from")
	readonly city: string;

	constructor(firstName: string, lastName: string, city: string, company: string) {
		this.firstName = firstName;
		this.lastName = lastName;
		this.city = city;
		this.company = company;
	}
}


/**
 * A simple feed providing a single collection of "visitors" loaded from a local
 * CSV file.
 */
export class Visitors1 extends feed.Feed {
	constructor(env: feed.FeedEnv) {
		super(env);
		this.establishFeed(new Collection("visitors"));
	}
}

/**
 * My single collection. A feed may provide several (named) collections, each one
 * with the same basic structure, but containing different data sets. Furthermore,
 * a collection may provide brief list data and detailed item data. But in this
 * case, the data is so simple and cheap, so we're providing the same data for
 * both lists and details.
 */
class Collection implements feed.StaticFeed<Item, Item> {
	readonly name: string;		// Our internal (short) name of this feed
	readonly listType = Item;	// My list and details provide the same type of items
	readonly itemType = Item;	// Yup, just said that

	private items: Item[];		// List fetched most recently, also used by getDetails

	constructor(feedName: string) {
		this.name = feedName;
	}

	/**
	 * Get a list of items from the spreadsheet.
	 */
	getList(spec:feed.FeedListSpec): Promise<ListData<Item>> {
		return SimpleFile.readCsv("CSVList.csv").then(data => (
			{ // The promised ListData with its items
				items: this.gotData(data)
			}
		));
	}

	/**
	 * Repackage data loaded from CSV file into what I'm supposed to return
	 */
	private gotData(csvList: CSVData[]): Item[] {
		const result: Item[] = [];

		for (const csvItem of csvList) {
			result.push(new Item(
				csvItem.firstName,
				csvItem.lastName,
				csvItem.city,
				csvItem.company
			));
		};
		this.items = result;	// Hold on to those for getDetails use
		return result;
	}

	/* Since all data is provided by the getList call, we don't expect any
	 * getDetails call, but need to implement one anyway to make everyone happy.
	 */
	getDetails(spec: FeedDetailsSpec): Promise<Item | undefined> {
		return Promise.resolve(this.items[spec.index]);
	}
}
