/*	A feed script loading all data from a local CSV file, including the time of each event.
	This is veri similar to the more basic Visitors1, but provide additional data and functionality.
	I weed out entries clearly in the past, and mark events currently "in session" so they
	can get different styling.

 	Copyright (c) 2021 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
 */

import * as feed from "system_lib/Feed";
import {SimpleFile} from "system/SimpleFile";
import {FeedDetailsSpec, FeedListSpec, ListData} from "system_lib/Feed";
import {field} from "system_lib/Metadata";

/**
 * Data read from CSV file.
 */
interface CSVData {
	startTime: string;
	endTime: string;
	firstName: string;
	lastName: string;
	city: string;
	company: string;
}

/**
 * Data provided to Blocks describing each feed item.
 */
class Item {

	@field("Event start time")
	readonly starts: string;

	@field("Event has started")
	readonly inProgress: boolean;

	@field("Given name")
	readonly firstName: string;

	@field("Surname")
	readonly lastName: string;

	@field("The company the fisitor work for")
	readonly company: string;

	@field("The city the visitor is from")
	readonly city: string;

	constructor(
			starts: Date, inProgress: boolean,
			firstName: string, lastName: string,
			city: string, company: string
	) {
		this.starts = dateToTimeOfDay(starts);
		this.inProgress = inProgress;
		this.firstName = firstName;
		this.lastName = lastName;
		this.city = city;
		this.company = company;
	}
}

/**	Return the time of day from date, in 24 hour HH:MM format.
*/
function dateToTimeOfDay(date: Date): string {
	return twoDigits(date.getHours()) + ':' + twoDigits(date.getMinutes());
}

/**	If num is a singl digit, prepend a 0 to make it two digits.
*/
function twoDigits(num: number): string {
	let result = num.toString();
	if (result.length === 1)
		result = '0' + result;
	return result;
}

/**	Interpret timeStr as "HH:MM", and apply that time to the
	date object, then return the date object.
*/
function setTime(date: Date, timeStr: string): Date {
	const parts = timeStr.split(':');

	const hours = parseInt(parts[0]);
	const minutes = parseInt(parts[1]);

	if (isNaN(hours) || hours < 0 || hours > 23)
		console.warn("Invalid hours", timeStr);
	else
		date.setHours(hours);
	if (isNaN(minutes) || minutes < 0 || minutes > 59)
		console.warn("Invalid miutes", timeStr);
	else
		date.setMinutes(minutes);

	date.setSeconds(0);
	date.setMilliseconds(0);
	return date;
}

/**
 * A simple feed providing a single collection of "visitors" loaded from a local
 * CSV file.
 */
export class Visitors2 extends feed.Feed {
	constructor(env: feed.FeedEnv) {
		super(env);
		this.establishFeed(new Collection("visitors"));
	}
}

/**
 * My single collection. A feed may provide several (named) collections, each one
 * with the same basic structure, but containing different data sets. Furthermore,
 * a collection may provide brief list data and detailed item data. But in this
 * case, the data is so simple and cheap, so we're providing the same data for
 * both lists and details.
 */
class Collection implements feed.StaticFeed<Item, Item> {
	readonly name: string;	// Our internal (short) name of this feed
	readonly listType = Item;	// My list and details provide the same type of items
	readonly itemType = Item;	// Yup, just said that

	private items: Item[];		// List fetched most recently, also used by getDetails

	constructor(feedName: string) {
		this.name = feedName;
	}

	/**
	 * Get a list of items from the spreadsheet.
	 */
	getList(spec:feed.FeedListSpec): Promise<ListData<Item>> {
		const fileName = "CSVList.csv";
		return SimpleFile.readCsv(fileName).then(data => (
			{ // The promised ListData with its items
				items: this.gotData(data)
			}
		), error => console.error(fileName, "not found", error));
	}

	/**
	 * Repackage data loaded from CSV file into what I'm supposed to return
	 */
	private gotData(csvList: CSVData[]): Item[] {
		const result: Item[] = [];
		const now = new Date();
		// console.log("Now", now.toString());
		for (const csvItem of csvList) {
			const nowTime = now.getTime();
			const startTime = setTime(new Date(nowTime), csvItem.startTime);
			// console.log("Start", startTime.toString(), csvItem.lastName);
			const endTime = setTime(new Date(nowTime), csvItem.endTime);
			console.log("End", endTime.toString(), csvItem.lastName);

			if (endTime.getTime() > nowTime) { // Event still in range
				result.push(new Item(
					startTime,
					startTime.getTime() < nowTime,
					csvItem.firstName,
					csvItem.lastName,
					csvItem.city,
					csvItem.company
				));
			} // Else events end time is in the past - skip it
		};
		this.items = result;	// Hold on to those for getDetails use
		return result;
	}

	/* Since all data is provided by the getList call, we don't expect any
	 * getDetails call, but need to implement one anyway to make everyone happy.
	 */
	getDetails(spec: FeedDetailsSpec): Promise<Item | undefined> {
		return Promise.resolve(this.items[spec.index]);
	}
}
